/*************************************************************************/
/* Copyright (C) 2007,2008 sacamano <m.sacamano@gmail.com>		 */
/* 									 */
/* This program is free software: you can redistribute it and/or modify	 */
/* it under the terms of the GNU General Public License as published by	 */
/* the Free Software Foundation, either version 3 of the License, or	 */
/* (at your option) any later version.					 */
/* 									 */
/* This program is distributed in the hope that it will be useful,	 */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of	 */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	 */
/* GNU General Public License for more details.				 */
/* 									 */
/* You should have received a copy of the GNU General Public License	 */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*************************************************************************/

#include "consonance.h"

/* Handler for 'version' option on the cmdline */

gboolean cmd_version(const gchar *opt_name,
		     const gchar *val,
		     struct con_win *cwin,
		     GError **error)
{
	g_print("Consonance %s\n", PACKAGE_VERSION);
	exit(0);
}

gboolean cmd_play(const gchar *opt_name,
		  const gchar *val,
		  struct con_win *cwin,
		  GError **error)
{
	if (!cwin->cstate->unique_instance) {
		dbus_send_signal(DBUS_SIG_PLAY, cwin);
		exit(0);
	}

	return TRUE;
}

gboolean cmd_stop(const gchar *opt_name,
		  const gchar *val,
		  struct con_win *cwin,
		  GError **error)
{
	if (!cwin->cstate->unique_instance) {
		dbus_send_signal(DBUS_SIG_STOP, cwin);
		exit(0);
	}

	return TRUE;
}

gboolean cmd_pause(const gchar *opt_name,
		   const gchar *val,
		   struct con_win *cwin,
		   GError **error)
{
	if (!cwin->cstate->unique_instance) {
		dbus_send_signal(DBUS_SIG_PAUSE, cwin);
		exit(0);
	}

	return TRUE;
}

gboolean cmd_prev(const gchar *opt_name,
		  const gchar *val,
		  struct con_win *cwin,
		  GError **error)
{
	if (!cwin->cstate->unique_instance) {
		dbus_send_signal(DBUS_SIG_PREV, cwin);
		exit(0);
	}

	return TRUE;
}

gboolean cmd_next(const gchar *opt_name,
		  const gchar *val,
		  struct con_win *cwin,
		  GError **error)
{
	if (!cwin->cstate->unique_instance) {
		dbus_send_signal(DBUS_SIG_NEXT, cwin);
		exit(0);
	}

	return TRUE;
}

gboolean cmd_inc_volume(const gchar *opt_name,
			const gchar *val,
			struct con_win *cwin,
			GError **error)
{
	if (!cwin->cstate->unique_instance) {
		dbus_send_signal(DBUS_SIG_INC_VOL, cwin);
		exit(0);
	}

	return TRUE;
}

gboolean cmd_dec_volume(const gchar *opt_name,
			const gchar *val,
			struct con_win *cwin,
			GError **error)
{
	if (!cwin->cstate->unique_instance) {
		dbus_send_signal(DBUS_SIG_DEC_VOL, cwin);
		exit(0);
	}

	return TRUE;
}

gboolean cmd_show_osd(const gchar *opt_name,
		      const gchar *val,
		      struct con_win *cwin,
		      GError **error)
{
	if (!cwin->cstate->unique_instance) {
		dbus_send_signal(DBUS_SIG_SHOW_OSD, cwin);
		exit(0);
	}

	return TRUE;
}

gboolean cmd_add_file(const gchar *opt_name,
		      const gchar *val,
		      struct con_win *cwin,
		      GError **error)
{
	gboolean ret = TRUE;
	DBusMessage *msg = NULL;

	msg = dbus_message_new_signal(DBUS_PATH, DBUS_INTERFACE, DBUS_SIG_ADD_FILE);
	if (!msg) {
		g_critical("(%s): Unable to allocate memory for DBUS message", __func__);
		ret = FALSE;
		goto exit;
	}
	dbus_message_append_args(msg, DBUS_TYPE_STRING, &val, DBUS_TYPE_INVALID);

	if (!dbus_connection_send(cwin->con_dbus, msg, NULL)) {
		g_critical("(%s): Unable to send DBUS message", __func__);
		ret = FALSE;
		goto bad;
	}

	dbus_connection_flush(cwin->con_dbus);
bad:
	dbus_message_unref(msg);
exit:
	return ret;
}
