/*************************************************************************/
/* Copyright (C) 2007,2008 sacamano <m.sacamano@gmail.com>		 */
/* 									 */
/* This program is free software: you can redistribute it and/or modify	 */
/* it under the terms of the GNU General Public License as published by	 */
/* the Free Software Foundation, either version 3 of the License, or	 */
/* (at your option) any later version.					 */
/* 									 */
/* This program is distributed in the hope that it will be useful,	 */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of	 */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	 */
/* GNU General Public License for more details.				 */
/* 									 */
/* You should have received a copy of the GNU General Public License	 */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*************************************************************************/

#include "consonance.h"

/* Handler for the preferences dialog */

static void pref_dialog_cb(GtkDialog *dialog, gint response_id, struct con_win *cwin)
{
	gboolean hf, aa, osd, ret;
	gchar *folder = NULL;
	GtkTreeIter iter;
	GtkTreeModel *model;

	switch(response_id) {
	case GTK_RESPONSE_CANCEL:
		break;
	case GTK_RESPONSE_OK:

		/* Hidden files */

		hf = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(cwin->cpref->hidden_files));
		if (hf)
			cwin->cpref->show_hidden_files = TRUE;
		else
			cwin->cpref->show_hidden_files = FALSE;

		/* Album art */

		aa = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(cwin->cpref->album_art));
		if (aa)
			cwin->cpref->show_album_art = TRUE;
		else
			cwin->cpref->show_album_art = FALSE;

		album_art_toggle_state(cwin);

		/* OSD */

		osd = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(cwin->cpref->osd));
		if (osd) {
			cwin->cpref->show_osd = TRUE;
			if (!notify_is_initted()) {
				if (!notify_init(PACKAGE_NAME))
					cwin->cpref->show_osd = FALSE;
			}
		}
		else
			cwin->cpref->show_osd = FALSE;

		/* Library */

		model = gtk_tree_view_get_model(GTK_TREE_VIEW(cwin->cpref->library_view));
		ret = gtk_tree_model_get_iter_first(model, &iter);

		/* Free the list of libraries and rebuild it again */

		free_library_dir(cwin);

		while (ret) {
			gtk_tree_model_get(model, &iter, 0, &folder, -1);
			if (folder)
				cwin->cpref->library_dir = g_slist_append(cwin->cpref->library_dir, folder);
			ret = gtk_tree_model_iter_next(model, &iter);
		}

		save_preferences(cwin);

		break;
	default:
		break;
	}

	gtk_widget_destroy(GTK_WIDGET(dialog));
	cwin->cpref->hidden_files = NULL;
	cwin->cpref->library_view = NULL;
}

/* Handler for adding a new library */

static void library_add_cb(GtkButton *button, struct con_win *cwin)
{
	GtkWidget *dialog;
	gint resp;
	gchar *folder;
	GtkTreeIter iter;
	GtkTreeModel *model;

	/* Create a folder chooser dialog */

	dialog = gtk_file_chooser_dialog_new("Select a folder to add to library",
					     GTK_WINDOW(cwin->mainwindow),
					     GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER,
					     GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
					     GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
					     NULL);

	/* Show it and get the folder */

	resp = gtk_dialog_run(GTK_DIALOG(dialog));

	switch (resp) {
	case GTK_RESPONSE_ACCEPT:
		model = gtk_tree_view_get_model(GTK_TREE_VIEW(cwin->cpref->library_view));
		folder = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
		if (folder) {
			gtk_list_store_append(GTK_LIST_STORE(model), &iter);
			gtk_list_store_set(GTK_LIST_STORE(model), &iter, 0, folder, -1);
			g_free(folder);
		}
		break;
	default:
		break;
	}

	gtk_widget_destroy(GTK_WIDGET(dialog));
}

/* Handler for removing a library */

static void library_remove_cb(GtkButton *button, struct con_win *cwin)
{
	GtkTreeSelection *selection;
	GtkTreeModel *model;
	GtkTreeIter iter;

	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(cwin->cpref->library_view));
	if (gtk_tree_selection_get_selected(selection, &model, &iter))
		gtk_list_store_remove(GTK_LIST_STORE(model), &iter);
}

static void update_preferences(struct con_win *cwin)
{
	gint cnt = 0, i;
	GSList *list;
	GtkTreeIter iter;
	GtkTreeModel *model;

	/* Update Hidden Files */

	if (cwin->cpref->show_hidden_files)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cwin->cpref->hidden_files), TRUE);

	/* Update album art */

	if (cwin->cpref->show_album_art)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cwin->cpref->album_art), TRUE);

	/* Update OSD */

	if (cwin->cpref->show_osd)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(cwin->cpref->osd), TRUE);

	/* Append libraries, if any */

	if (cwin->cpref->library_dir) {
		model = gtk_tree_view_get_model(GTK_TREE_VIEW(cwin->cpref->library_view));

		cnt = g_slist_length(cwin->cpref->library_dir);
		list = cwin->cpref->library_dir;

		for (i=0; i<cnt; i++) {
			gtk_list_store_append(GTK_LIST_STORE(model), &iter);
			gtk_list_store_set(GTK_LIST_STORE(model), &iter, 0, (gchar*)list->data, -1);
			list = list->next;
		}
	}
}

void save_preferences(struct con_win *cwin)
{
	const gchar *col_name;
	gchar *data, **libs, **columns, **nodes;
	gsize length;
	gint cnt = 0, i = 0, *col_widths, *window_size;
	gint win_width, win_height;
	GError *error = NULL;
	GSList *list;
	GList *cols, *j;
	GtkTreeViewColumn *col;

	/* Version */

	g_key_file_set_string(cwin->cpref->configrc_keyfile,
			      GROUP_GENERAL,
			      KEY_INSTALLED_VERSION,
			      PACKAGE_VERSION);

	/* Shuffle and repeat options */

	g_key_file_set_boolean(cwin->cpref->configrc_keyfile,
			       GROUP_GENERAL,
			       KEY_SHUFFLE,
			       cwin->cpref->shuffle);
	g_key_file_set_boolean(cwin->cpref->configrc_keyfile,
			       GROUP_GENERAL,
			       KEY_REPEAT,
			       cwin->cpref->repeat);

	/* Library view order */

	g_key_file_set_integer(cwin->cpref->configrc_keyfile,
			       GROUP_LIBRARY,
			       KEY_LIBRARY_VIEW_ORDER,
			       cwin->cpref->cur_library_view);

	/* Hidden Files option */

	g_key_file_set_boolean(cwin->cpref->configrc_keyfile,
			       GROUP_GENERAL,
			       KEY_SHOW_HIDDEN_FILE,
			       cwin->cpref->show_hidden_files);

	/* Album art option */

	g_key_file_set_boolean(cwin->cpref->configrc_keyfile,
			       GROUP_GENERAL,
			       KEY_SHOW_ALBUM_ART,
			       cwin->cpref->show_album_art);

	/* OSD option */

	g_key_file_set_boolean(cwin->cpref->configrc_keyfile,
			       GROUP_GENERAL,
			       KEY_SHOW_OSD,
			       cwin->cpref->show_osd);

	/* Window size */

	window_size = g_new0(gint, 2);
	gtk_window_get_size(GTK_WINDOW(cwin->mainwindow),
			    &win_width,
			    &win_height);
	window_size[0] = win_width;
	window_size[1] = win_height;

	g_key_file_set_integer_list(cwin->cpref->configrc_keyfile,
				    GROUP_GENERAL,
				    KEY_WINDOW_SIZE,
				    window_size,
				    2);
	g_free(window_size);

	/* List of libraries */

	if (cwin->cpref->library_dir) {
		list = cwin->cpref->library_dir;
		cnt = g_slist_length(cwin->cpref->library_dir);
		libs = g_new0(gchar *, cnt);

		for (i=0; i<cnt; i++) {
			libs[i] = (gchar*)list->data;
			list = list->next;
		}

		g_key_file_set_string_list(cwin->cpref->configrc_keyfile,
					   GROUP_LIBRARY,
					   KEY_LIBRARY_DIR,
					   (const gchar **)libs,
					   cnt);
		g_free(libs);
	}
	else {
		if (g_key_file_has_group(cwin->cpref->configrc_keyfile, GROUP_LIBRARY) && 
		    g_key_file_has_key(cwin->cpref->configrc_keyfile,
				       GROUP_LIBRARY,
				       KEY_LIBRARY_DIR,
				       &error)) {
			g_key_file_remove_key(cwin->cpref->configrc_keyfile,
					      GROUP_LIBRARY,
					      KEY_LIBRARY_DIR,
					      &error);
		}
	}

	/* List of columns visible in current playlist */

	if (cwin->cpref->playlist_columns) {
		list = cwin->cpref->playlist_columns;
		cnt = g_slist_length(cwin->cpref->playlist_columns);
		columns = g_new0(gchar *, cnt);

		for (i=0; i<cnt; i++) {
			columns[i] = (gchar*)list->data;
			list = list->next;
		}

		g_key_file_set_string_list(cwin->cpref->configrc_keyfile,
					   GROUP_GENERAL,
					   KEY_PLAYLIST_COLUMNS,
					   (const gchar **)columns,
					   cnt);
		g_free(columns);
	}

	/* Column widths */

	cols = gtk_tree_view_get_columns(GTK_TREE_VIEW(cwin->current_playlist));
	cnt = g_list_length(cols);
	if (cols) {
		col_widths = g_new0(gint, cnt);
		for (j=cols, i=0; j != NULL; j = j->next) {
			col = j->data;
			col_name = gtk_tree_view_column_get_title(col);
			if (is_present_str_list(col_name, cwin->cpref->playlist_columns))
				col_widths[i++] = gtk_tree_view_column_get_width(col);
		}
		g_key_file_set_integer_list(cwin->cpref->configrc_keyfile,
					    GROUP_GENERAL,
					    KEY_PLAYLIST_COLUMN_WIDTHS,
					    col_widths,
					    i);
		g_list_free(cols);
		g_free(col_widths);
	}

	/* Library tree nodes */

	if (cwin->cpref->library_tree_nodes) {
		list = cwin->cpref->library_tree_nodes;
		cnt = g_slist_length(cwin->cpref->library_tree_nodes);
		nodes = g_new0(gchar *, cnt);

		for (i=0; i<cnt; i++) {
			nodes[i] = (gchar*)list->data;
			list = list->next;
		}

		g_key_file_set_string_list(cwin->cpref->configrc_keyfile,
					   GROUP_LIBRARY,
					   KEY_LIBRARY_TREE_NODES,
					   (const gchar **)nodes,
					   cnt);
		g_free(nodes);
	}

	/* Save to conrc */

	data = g_key_file_to_data(cwin->cpref->configrc_keyfile, &length, &error);
	if (!g_file_set_contents(cwin->cpref->configrc_file, data, length, &error))
		g_critical("(%s): Unable to write preferences file : %s", __func__, error->message);

	g_free(data);
}

void preferences_dialog(struct con_win *cwin)
{
	GtkWidget *vbox_general, *hbox_library;
	GtkWidget *dialog, *label_general, *label_library;
	GtkWidget *hidden_align, *album_align, *osd_align, *library_align;
	GtkWidget *hidden_files, *album_art, *osd, *library_view;
	GtkWidget *library_bbox_align, *library_bbox, *library_add, *library_remove;
	GtkListStore *library_store;
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;

	/* The main preferences dialog */

	dialog = gtk_dialog_new_with_buttons("Preferences",
					     GTK_WINDOW(cwin->mainwindow),
					     GTK_DIALOG_MODAL,
					     GTK_STOCK_CANCEL,
					     GTK_RESPONSE_CANCEL,
					     GTK_STOCK_OK,
					     GTK_RESPONSE_OK,
					     NULL);

	vbox_general = gtk_vbox_new(FALSE, 2);

	/* Labels */

	label_general = gtk_label_new(NULL);
	gtk_label_set_markup(GTK_LABEL(label_general), "<b>General</b>");
	gtk_misc_set_alignment(GTK_MISC(label_general), 0, 0);

	label_library = gtk_label_new(NULL);
	gtk_label_set_markup(GTK_LABEL(label_library), "<b>Library</b>");
	gtk_misc_set_alignment(GTK_MISC(label_library), 0, 0);

	/* Hidden files */

	hidden_align = gtk_alignment_new(0,0,0,0);
	gtk_alignment_set_padding(GTK_ALIGNMENT(hidden_align), 0, 0, 20, 0);
	hidden_files = gtk_check_button_new_with_label("Show Hidden Files in File View");
	gtk_container_add(GTK_CONTAINER(hidden_align), hidden_files);

	/* Album art */

	album_align = gtk_alignment_new(0,0,0,0);
	gtk_alignment_set_padding(GTK_ALIGNMENT(album_align), 0, 0, 20, 0);
	album_art = gtk_check_button_new_with_label("Show Album art in Panel");
	gtk_container_add(GTK_CONTAINER(album_align), album_art);

	/* Notification */

	osd_align = gtk_alignment_new(0,0,0,0);
	gtk_alignment_set_padding(GTK_ALIGNMENT(osd_align), 0, 0, 20, 0);
	osd = gtk_check_button_new_with_label("Show OSD for track change");
	gtk_container_add(GTK_CONTAINER(osd_align), osd);

 	/* Library List */

	hbox_library= gtk_hbox_new(FALSE, 2);
	library_align = gtk_alignment_new(0,0,0,0);
	gtk_alignment_set_padding(GTK_ALIGNMENT(library_align), 0, 0, 20, 0);

	library_store = gtk_list_store_new(1, G_TYPE_STRING);
	library_view = gtk_tree_view_new_with_model(GTK_TREE_MODEL(library_store));

	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes("Folder", renderer, "text", 0, NULL);
	gtk_tree_view_column_set_resizable(column, GTK_TREE_VIEW_COLUMN_GROW_ONLY);
	gtk_tree_view_append_column(GTK_TREE_VIEW(library_view), column);

	library_bbox_align = gtk_alignment_new(0, 0.5, 0, 0);
	library_bbox = gtk_vbutton_box_new();
	library_add = gtk_button_new_from_stock(GTK_STOCK_ADD);
	library_remove = gtk_button_new_from_stock(GTK_STOCK_REMOVE);

	gtk_box_pack_start(GTK_BOX(library_bbox),
			   library_add,
			   FALSE,
			   FALSE,
			   0);
	gtk_box_pack_start(GTK_BOX(library_bbox),
			   library_remove,
			   FALSE,
			   FALSE,
			   0);

	gtk_container_add(GTK_CONTAINER(library_bbox_align), library_bbox);

	/* Pack all library category items */

	gtk_box_pack_start(GTK_BOX(hbox_library),
			   library_view,
			   TRUE,
			   FALSE,
			   0);
	gtk_box_pack_start(GTK_BOX(hbox_library),
			   library_bbox_align,
			   TRUE,
			   FALSE,
			   0);

	gtk_container_add(GTK_CONTAINER(library_align), hbox_library);

	/* Pack all items */

	gtk_box_pack_start(GTK_BOX(vbox_general),
			   label_general,
			   FALSE,
			   FALSE,
			   0);
	gtk_box_pack_start(GTK_BOX(vbox_general),
			   hidden_align,
			   FALSE,
			   FALSE,
			   0);
	gtk_box_pack_start(GTK_BOX(vbox_general),
			   album_align,
			   FALSE,
			   FALSE,
			   0);
	gtk_box_pack_start(GTK_BOX(vbox_general),
			   osd_align,
			   FALSE,
			   FALSE,
			   0);
	gtk_box_pack_start(GTK_BOX(vbox_general),
			   label_library,
			   FALSE,
			   FALSE,
			   0);
	gtk_box_pack_start(GTK_BOX(vbox_general),
			   library_align,
			   FALSE,
			   FALSE,
			   0);

	/* Add to dialog */

	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), vbox_general);

	/* Store references */

	cwin->cpref->hidden_files = hidden_files;
	cwin->cpref->album_art = album_art;
	cwin->cpref->osd = osd;
	cwin->cpref->library_view = library_view;

	/* Setup signal handlers */

	g_signal_connect(G_OBJECT(dialog), "response",
			 G_CALLBACK(pref_dialog_cb), cwin);
	g_signal_connect(G_OBJECT(library_add), "clicked",
			 G_CALLBACK(library_add_cb), cwin);
	g_signal_connect(G_OBJECT(library_remove), "clicked",
			 G_CALLBACK(library_remove_cb), cwin);

	update_preferences(cwin);

	gtk_widget_show_all(dialog);
}

void free_library_dir(struct con_win *cwin)
{
	gint cnt = 0, i;
	GSList *list;

	list = cwin->cpref->library_dir;
	cnt = g_slist_length(cwin->cpref->library_dir);

	for (i=0; i<cnt; i++) {
		g_free(list->data);
		list = list->next;
	}

	g_slist_free(cwin->cpref->library_dir);
	cwin->cpref->library_dir = NULL;
}

void free_playlist_columns(struct con_win *cwin)
{
	gint cnt = 0, i;
	GSList *list;

	list = cwin->cpref->playlist_columns;
	cnt = g_slist_length(cwin->cpref->playlist_columns);

	for (i=0; i<cnt; i++) {
		g_free(list->data);
		list = list->next;
	}

	g_slist_free(cwin->cpref->playlist_columns);
	cwin->cpref->playlist_columns = NULL;
}

void free_library_tree_nodes(struct con_win *cwin)
{
	gint cnt = 0, i;
	GSList *list;

	list = cwin->cpref->library_tree_nodes;
	cnt = g_slist_length(cwin->cpref->library_tree_nodes);

	for (i=0; i<cnt; i++) {
		g_free(list->data);
		list = list->next;
	}

	g_slist_free(cwin->cpref->library_tree_nodes);
	cwin->cpref->library_tree_nodes = NULL;
}
